
import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.DataInputStream;
import java.io.DataOutputStream;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.OutputStreamWriter;
import java.nio.ByteBuffer;
import java.nio.ByteOrder;
import java.util.ArrayList;
import java.util.Scanner;


import Jama.Matrix;
import Jama.EigenvalueDecomposition;

public class tSne {
	public static int checkpoint = 1;
	public static void printCheckpoint(String s){
		System.out.println("\n[Stage "+checkpoint+"]: " + s);
		checkpoint++;
	}
	
	public static void main (String []args){
		System.out.println("Visgenex (Java Version): v1.0");
		System.out.print("    Uses the JAMA Java Libraries:\n     http://math.nist.gov/javanumerics/jama/\n");
		System.out.println("    Uses t-SNE Algorithm t-SNE home page:\n     http://homepage.tudelft.nl/19j49/t-SNE.html\n");
		String inputFileName =  "input.csv";
		String outputFileName = null;
		boolean headers = true;
		double perplexity = 30;
		String inputProcessing = null;
		int groupSize = 1;
		int type = 1; // 0 = raw values, 1 = z-scores, 2 = zero-means, 3= fold-induction, 4=percent-induction
		if (args.length < 1){
			System.out.println(" Parameter List:");
			System.out.println("   -i <filename> - set the csv input filename");
			System.out.println("      default: 'input.csv' ");
			System.out.println("   -h <y/n> - is the first row of the input file column headers instead of data?");
			System.out.println("      default: y");
			System.out.println("   -o <filename> - set the csv output filename");
			System.out.println("      default: '<inputfilename>_output.csv' ");
			System.out.println("   -p <integer> - perplexity value");
			System.out.println("      default: 30 ");
			System.out.println("   -gs <integer> - number of replicates per condition, groups adjacent columns into groups of <integer> and takes their mean.");
			System.out.println("      default: 1 ");
			System.out.println("   -ip <text> - input processing");
			System.out.println("      default: 'z-score'");
			System.out.println("      options: 'raw' - unprocessed data");
			System.out.println("               'z-score' - using z-scores");
			System.out.println("               'zero-means' - subtracts the row average from each row, ");
			System.out.println("               'fold-induction' - divides entire row by its lowest value ");
			System.out.println("               'percent-induction' - subtract entire row by lowest value in the row and then divides by the highest value (scales entire row (0=<n=<1)");
			System.out.println("");
			System.out.println(" No Parameters specified, continue using defaults (y/n)?");
			Scanner keyboard = new Scanner(System.in);
			if (keyboard.next().toLowerCase().contains("n")) {
				return;
			}
		} else {
			for (int i = 0; i < args.length-1; i++){
				if (args[i].compareTo("-o")== 0){
					outputFileName = args[i+1];
					i++;
				}
				else if (args[i].compareTo("-i")== 0){
					inputFileName = args[i+1];
					i++;
				}
				else if (args[i].compareTo("-p")==0){
					perplexity = Double.parseDouble(args[i+1]);
					i++;
				}
				else if (args[i].compareTo("-gs")==0){
					groupSize = Integer.parseInt(args[i+1]);
					i++;
				}
				else if (args[i].compareTo("-h")==0){
					headers = args[i+1].toLowerCase().contains("y");
					i++;
				}
				else if (args[i].compareTo("-ip")==0){
					inputProcessing= args[i+1];
					i++;
				}
			}
		}
		
		//inputFileName.split(".").get(0);
		if ( outputFileName == null) {
			String [] blocks = inputFileName.split("\\.");
			String newname = "";
			System.out.println("inputFileName " + inputFileName + " blocks " + blocks.length);
			for (int i =0; i < blocks.length-1; i++){
				if (i > 0 )newname = newname + ".";
				newname = newname + blocks[i];
			}
			outputFileName= newname +"_output.csv";
		}
		// 0 = raw values, 1 = z-scores
		// 0 = raw values, 1 = z-scores, 2 = zero-means, 3= fold-induction, 4=percent-induction
		if (inputProcessing == null) inputProcessing = "z-score"; // default
		if (inputProcessing.contains("raw")) {
			inputProcessing = "raw";
			type = 0; // raw
		} else if (inputProcessing.toLowerCase().contains("zero")) {
			inputProcessing = "zero-means";
			type = 2; // zero means
		} else if (inputProcessing.toLowerCase().contains("fold")) {
			inputProcessing = "fold-induction";
			type = 3; // fold induction
		} else if (inputProcessing.toLowerCase().contains("per")) {
			inputProcessing = "percent-induction";
			type = 4; //percent_induction
		} else {
			type = 1; // z-scores
			inputProcessing = "z-score";
		}
		
 
		System.out.println(" Parameters as Follows:");
		System.out.println("   input  file: " + inputFileName);
		System.out.println("   output file: " + outputFileName);
		System.out.println("   perplexity: " + perplexity);
		System.out.println("   input processing: " + inputProcessing);
		System.out.println("   header row in input file: " + (headers ?  "yes" : "no"));
		System.out.println("   group size: " + groupSize);
		//System.out.println("\n");
		
		/**** 
		 * PARSE INPUT FILE 
		 ****/
		printCheckpoint("Parsing Input File...");
		ArrayList<String> input= fileToStrings(inputFileName);
		if (input.size() < 1) {
			System.out.println("\nERROR INPUT FILE: Missing or empty.\n");
			return;
		}
		int columns = input.get(0).split(",").length-1; // first row is names
		int rowOffset = 0;
		if (headers) rowOffset = 1;
		int rows = input.size()-rowOffset;
		double rawNumbers[][] = new double[rows][columns]; // Matlab style
		String rowIds[] = new String[rows];
		/*@SuppressWarnings("unused")
		String colNames[] = new String[columns];
		colNames = input.get(0).split(",");
		*/
		
		// loop through rows, parsing Numbers
		for (int i = 0; i < rows ; i++){
			String [] thisRow = input.get(i+rowOffset).split(",");
			rowIds[i] = thisRow[0];
			for (int j = 0; j < columns; j++){
				try {rawNumbers[i][j] = Double.parseDouble(thisRow[j+1]); } catch(Exception e) { rawNumbers[i][j] = 0;}
			}
		}
		
		// group averages
		if (columns%groupSize != 0){
			System.out.println("\nERROR: Group size parameter is incorrect.\n");
			return;
		}
		if (groupSize != 1){ // do nothing if 1
			int newCols = columns/groupSize;
			double groupedNums[][] = new double[rows][newCols];
			for (int i = 0; i < newCols; i++){
				for (int j = 0; j < rows; j++){
					double avg = 0;
					for (int k = i*groupSize; k < (i+1)*groupSize ;k++) 
						avg += rawNumbers[j][k];
					avg = avg / groupSize;
					groupedNums[j][i] = avg;
				}
			}
			rawNumbers = groupedNums;
			columns = rawNumbers[0].length;
			rows = rawNumbers.length;
		}
		
		
		
		
		//Sorting them? May be required
		/****
		 * PreProcesssing
		 */
		// 0 = raw values, 1 = z-scores, 2 = zero-means, 3= fold-induction, 4=percent-induction
		if (type == 0) {
				// Do nothing keep raw values
		} else if (type == 1) { // Z Scores...
			double [][] zScores = z_scores(rawNumbers);//rawNumbers);
			rawNumbers = zScores;
		} else if (type == 2) { // zero means
			for (int i = 0; i < rawNumbers.length; i++){
				// calculate row mean
				double total = 0;
				for (int j = 0; j < rawNumbers[i].length; j++){
					total += rawNumbers[i][j];
				}
				double mean = total / rawNumbers[i].length;
				for (int j = 0; j < rawNumbers[i].length; j++){
					rawNumbers[i][j] = rawNumbers[i][j] - mean;
				}
				// subtract from all
			}
			
		} else if (type == 3) { // fold induction
			/* Divide each row by lowest value */
				for (int i = 0; i < rawNumbers.length; i++){
					double lowest = Double.MAX_VALUE;
					for (int j = 0; j < rawNumbers[i].length; j++){
						if (lowest > rawNumbers[i][j]) lowest = rawNumbers[i][j];
					}
					for (int j = 0; j < rawNumbers[i].length; j++){
						rawNumbers[i][j] = rawNumbers[i][j] / lowest;
					}
					// subtract from all
				}
			
		} else if (type == 4) { // percent induction
			/* Subtract lowest value, then divide by max value (result everything 0<->1) */
			for (int i = 0; i < rawNumbers.length; i++){
				double lowest = Double.MAX_VALUE;
				for (int j = 0; j < rawNumbers[i].length; j++){
					if (lowest > rawNumbers[i][j]) lowest = rawNumbers[i][j];
				}
				for (int j = 0; j < rawNumbers[i].length; j++){
					rawNumbers[i][j] = rawNumbers[i][j] - lowest;
				}
				double highest = Double.MIN_VALUE;
				for (int j = 0; j < rawNumbers[i].length; j++){
					if (highest < rawNumbers[i][j]) highest = rawNumbers[i][j];
				}
				for (int j = 0; j < rawNumbers[i].length; j++){
					rawNumbers[i][j] = rawNumbers[i][j] / highest;
				}
				// subtract from all
			}
			
		}
		//printDoubleArray(rawNumbers);
		
		// PCA
		


		/**** 
		 * GENERATE t-SNE INPUT
		 ****/
		System.out.println(" generating t-SNE Instructions...");
		
		if (generateTsneInstructions(rawNumbers,2,perplexity) != 0) {
			System.out.println("\nERROR: Generating t-SNE Instructions.");
			return;
		}
		/**** 
		 * EXECUTE t-SNE
		 ****/
		printCheckpoint("Executing t-SNE with " + rows + " rows and " + columns + " columns...");
		System.out.println(" this stage can take quite some time depending on the size of your input data (possibly hours)");
		
		if (executeTsne() != 0) {
			System.out.println("\nERROR: running t-SNE executables.");
			return;
		}
		 /**** 
		  * RETRIEVE t-SNE Results
		  ****/
		 //Read t-SNE File
		printCheckpoint("Reading t-SNE Results...");
		String FILENAME2 = "result.dat";
			DataInputStream is;
			double tsneNumbers[][] = new double[rows][2]; 
			int landmarks[] = new int[rows];
			try {
				is = new DataInputStream(new FileInputStream( FILENAME2));
				//System.out.println("rows: " + rows + " columns: " + columns);
				int n = toByteInt(is.readInt());
				int d = toByteInt(is.readInt());
				System.out.println(" returned: " + n + "x" + d + " matrix");
				 for (int j = 0; j < rows; j++){
				  	for (int i = 0; i < 2; i++){
				  		tsneNumbers[j][i] = toByteDouble(is.readDouble());
				  	}
				  }
		 		 
				 for (int i = 0; i < rows; i++){ 
					 landmarks[i] = toByteInt(is.readInt());
				 }
			} catch (IOException e) {
				System.out.println("\nERROR: Reading t-SNE Result file");
				return;
			}
			
			
			/*for (int i = 0; i < rows; i++){
				System.out.println("land: " + landmarks[i]);
			}*/
			/**** 
			 * OUTPUT CSV
			 ****/
			try {
				System.out.println(" generating " + outputFileName + "...");
				FileOutputStream fo = new FileOutputStream(outputFileName);
				DataOutputStream out = new DataOutputStream(fo);
				BufferedWriter bw = new BufferedWriter(new OutputStreamWriter(out));
				 for (int i = 0; i < rows; i++){
					// System.out.println("i: " + i + " landmark: " + landmarks[i]);
					bw.write("\"" + rowIds[landmarks[i]].replace("\"","''") + "\"," + tsneNumbers[i][0] + "," +tsneNumbers[i][1] + "\n");
				 }
				 bw.close();
				 printCheckpoint("Complete: " + outputFileName + " written succesfully.");
				 //out.close();
			} catch (IOException e) {}

	}
	public static ArrayList<String> fileToStrings(String file){
		ArrayList<String> inputstrings = new ArrayList<String>();
		DataInputStream in = null;
	    try {
	    	FileInputStream fstream = new FileInputStream(file);
	    	  // Get the object of DataInputStream
	    	  in = new DataInputStream(fstream);
	    	  BufferedReader br = new BufferedReader(new InputStreamReader(in));	    	  
	    	  String strLine;
	    	  while ((strLine = br.readLine()) != null)   {
	    	  // Print the content on the console
	    		  inputstrings.add(strLine);
	    		//  System.out.println (strLine);
	    	  }
	    	
	    } catch(IOException e) { } finally {
	    	if (in != null) {
	    		try { in.close(); } catch (IOException e) { }
	        }
	    } 
	    return inputstrings;
	}
	public static int toByteInt(int i){
		ByteBuffer bb = ByteBuffer.allocate(Integer.SIZE/2);
	    bb.putInt(i);
	    bb.order(ByteOrder.LITTLE_ENDIAN);
	    return bb.getInt(0);
	}
	public static double toByteDouble(double d){
		ByteBuffer bb =ByteBuffer.allocate(Double.SIZE);		
		bb.putDouble(d);
		bb.order(ByteOrder.LITTLE_ENDIAN);
		return bb.getDouble(0);
	}
	public static int executeTsne() {
		String tSNE_Executable = "./tSNE_linux";
		if (System.getProperty("os.name").toLowerCase().contains("mac")){
			tSNE_Executable = "./tSNE_maci";
		} else if (System.getProperty("os.name").toLowerCase().contains("win")) {
			tSNE_Executable = "tSNE.exe";
		} 
		try {
			String s = null;
			//ProcessBuilder pb = new ProcessBuilder(tSNE_Executable);
			//pb.directory(new File(dir));
			//Process p = pb.start();
			Process p = Runtime.getRuntime().exec(tSNE_Executable);
			//BufferedReader stdInput = new BufferedReader(new 
			//InputStreamReader(p.getInputStream()));
			BufferedReader stdError = new BufferedReader(new 
					InputStreamReader(p.getErrorStream()));
			// while ((s = stdInput.readLine()) != null) {
			//     System.out.println(s);
			// }
		            
			while ((s = stdError.readLine()) != null) {
				System.out.println(s);
			}
	
		} catch (IOException e) {
			return -1;
		}	
		return 0;
	}
	public static double[][] returnPCA(double[][] initial){
		int columns = initial[0].length;
		int rows = initial.length;
		
		// suX = bsxfun(@minus, X, mean(X, 1));  (Zero meaning columns)
		//
		for (int i = 0; i < columns; i++){
			double total = 0;
			for (int j = 0; j < rows; j++){
				total += initial[j][i];
			}
			double mean = total / rows;
			for (int j = 0; j < rows; j++){
				initial[j][i] = initial[j][i] - mean;
			}
		} // A Rated

		//covX = X' * X;
		
		Matrix raw = new Matrix(initial);
		Matrix covX = raw.transpose().times(raw);
		EigenvalueDecomposition eig = new EigenvalueDecomposition(covX);
		// [M, lambda] = eig(covX);
		//EigenvalueDecomposition eig = covX.eig();
		System.out.println("    (PCA Calculation) Eigen Matrix Valid: " + (eigenTest(covX) ? "Yes" : "No"));
		
		Matrix M = eig.getV();//eig.getV(); // M CORRECT			

		// [lambda, ind] = sort(DIAG(LAMBDA), 'descend');		
		
		double eigenVals [][] = eig.getD().getArrayCopy();
		
		
		double [][]diaglambda = new double[	eigenVals.length][2]; // 2nd column = index
		for (int i = 0; i < 	diaglambda.length; i++){
			diaglambda[i][0] = 	eigenVals[i][i];
			diaglambda[i][1] = i;
		}	
		
		//[lambda, ind] = sort(diag(lambda), 'descend');
		for (int i = 0; i < diaglambda.length; i++){
			for (int j = diaglambda.length-1; j > i; j--) {
				if (diaglambda[j-1][0] < diaglambda[j][0]){ 
					double [] temp = new double[2]; 
					temp = diaglambda[j];
					diaglambda[j] = diaglambda[j-1];
					diaglambda[j-1] = temp;
				}
			}
		}

		// M = M(:,ind(1:initial_dims)); // Reorder M and M alone by the values
		double [][]dM = M.getArrayCopy();
		double [][]newsorted = new double[dM.length][dM[0].length];// M.getArrayCopy();
		for (int i = 0; i < diaglambda.length; i++){
			for (int j= 0; j < newsorted.length; j++) newsorted[j][i] = dM[j][(int)diaglambda[i][1]];
		}
		//coorrect here?
		
		//X = X * M; d

		
		M = raw.times( new Matrix(newsorted)); // This line is messed up? First Row/Last Column WRONG
		
		
		// return M
		return M.getArrayCopy();
		
	}
	
	public static void printDoubleArray(double mat1[][]){
		System.out.println("");
		for (int i = 0; i < mat1.length; i++){
			for (int j = 0; j < mat1[0].length; j++){
				System.out.printf(" %05f",mat1[i][j]);
			}
			System.out.println("");
		}
	}
	public static int generateTsneInstructions(double[][] inputMatrix,int dimensions, double perplexity){	
		inputMatrix =returnPCA(inputMatrix);	// fast_tsne in matlab does this

		String FILENAME = "data.dat";
	    DataOutputStream os;
		int columns = inputMatrix[0].length;
		int rows = inputMatrix.length;
		try {
			os = new DataOutputStream(new FileOutputStream( FILENAME));	
			
		    os.writeInt(toByteInt(rows)); 		//Row Count	    	    
		    os.writeInt(toByteInt(columns)); 	//Column Count	    
		    os.writeInt(toByteInt(2));			//dimensions to return	    
			os.writeDouble( toByteDouble(perplexity)); 	//perplexity
			os.writeDouble( toByteDouble(1)); 	//landmarks	
			
		   // Actual Data
			
			for (int i = 0; i < rows; i++){
				for (int j = 0; j < columns; j++){
		    		os.writeDouble( toByteDouble(inputMatrix[i][j]));	
				}
			}
		    
		    os.close();
		} catch (IOException execpt) {
			return -1;
		}
		return 0;
	}
	public static boolean eigenTest(Matrix m){
		EigenvalueDecomposition eig = new EigenvalueDecomposition(m);
		
		double [][] v = eig.getV().getArrayCopy();
		double[][]eigenVals = eig.getD().getArrayCopy();
		boolean same = true;
		for (int i = 0; i < eigenVals.length;i ++){
			double thisEigenColumn[][] = new double [v.length][1];
			for (int j = 0; j < thisEigenColumn.length; j++){
				thisEigenColumn[j][0] = v[j][i];
			}
			Matrix eigCol = new Matrix(thisEigenColumn);
			Matrix leftHand = m.times(eigCol);
			Matrix rightHand = eigCol.times( eigenVals[i][i]);

			double [][] dright = rightHand.getArrayCopy();
			double [][] dleft = leftHand.getArrayCopy();
			boolean thisRowSame = true;
			for (int j = 0; j < dright.length; j++){
				for (int k = 0; k <dright[0].length; k++){
					double t = (dleft[j][k]+2) / (dright[j][k]+2) ;
					if (!(t > 0.999999 && t < 1.000001)){
						thisRowSame = false;
					} 
				}	
			}
			if (!thisRowSame) same = false;
		}
		return same;
	}
	
	public static double [][] z_scores(double input[][]){

		// zmat = zero_mean_rows( mat ); 
		//      rowmeans = mean( mat, 2 ); 
		//      zmr = mat - repmat( rowmeans, [ 1 size(mat,2) ] ); 
		int rows = input.length;
		int columns = input[0].length;
		for (int i = 0; i < rows; i++){
			double total = 0;
			for (int j = 0; j < columns; j++){
				total += input[i][j];
			}
			double mean = total / columns;
			for (int j = 0; j < columns; j++){
				input[i][j] = input[i][j] - mean;
			}
		}
//		
		double [][]zmean = input;

		// zero mean rows

		//Std deviation code, works !
//		sds = std( zmat, 1, 2 );  COLUMN WISE
//		isds = 1 ./ ( sds + 1.0e-10 ); 
		double sds [] = new double[rows];
		for (int i = 0; i < rows;i++){ // through columns
			double sum = 0;
			double rowMean = 0;
			for (int j = 0; j < columns;j++) rowMean += input[i][j];
			rowMean /= columns;
			for (int j = 0; j < columns;j++) sum += (input[i][j] - rowMean)* (input[i][j] - rowMean);
			
			sds[i] = Math.sqrt(sum / columns) + 0.0000000001; //+ 1.0e-10 (Ensure not 0)
			sds[i] = 1 / sds[i]; //inverse isds
		}
//zmean
	
//
//		nsdm = zmat; 
//		for c = 1:size(nsdm,2)
//		    nsdm(:,c) = nsdm(:,c) .* isds; 
//		end
		for (int j = 0; j < columns; j++){
			for (int i =0; i < rows; i++){
				zmean[i][j] = zmean[i][j] * sds[i];
			}
		}
		
//		returning rsdm
		

		
		return input;
	}
}



